""" DataPipeline Base. """


import collections
import itertools


class DataPipelineBase(collections.Iterable):
    def __init__(self, **kwargs):
        self._items = dict()
        for key, val in kwargs.iteritems():
            setattr(self, key, val)

    def __iter__(self):
        raise NotImplementedError

    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError:
            return self._items[key]

    def __len__(self):
        return len(self._items)


class DataPipelineParameter(DataPipelineBase):
    def __init__(self, **attributes):
        assert 'id'    in attributes, "'id' not in attributes"
        assert 'value' in attributes, "'value' not in attributes"
        self.id    = attributes['id']
        self.value = attributes['value']
        del attributes['id']
        del attributes['value']
        super(DataPipelineParameter, self).__init__(**attributes)

    def __repr__(self):
        return "<%s id: \"%s\", value: \"%s\">" % \
            (type(self).__name__, self.id, self.value)

    def __setattr__(self, key, value):
        if key not in ('id', 'value', '_items'):
            self.attributes[key] = value
        super(DataPipelineParameter, self).__setattr__(key, value)

    def __iter__(self):
        iterhelper = lambda kv: { 'key' : kv[0], 'stringValue' : str(kv[-1]) }
        yield 'id',          self.id
        yield 'stringValue', self.value
        yield 'attributes',  list(map(iterhelper, self.attributes.iteritems()))

    @property
    def attributes(self):
        return self._items


class TypedDataPipelineParameter(DataPipelineParameter):
    def __init__(self, **attributes):
        type_ = getattr(self, 'TYPE_NAME', type(self).__name__)
        attributes.setdefault('type', type_)
        super(TypedDataPipelineParameter, self).__init__(**attributes)


class DataPipelineObject(DataPipelineBase):
    def __init__(self, **fields):
        assert 'id'   in fields, "'id' not in fields"
        assert 'name' in fields, "'value' not in fields"
        self.id   = fields['id']
        self.name = fields['name']
        del fields['id']
        del fields['name']
        super(DataPipelineObject, self).__init__(**fields)

    def __repr__(self):
        return "<%s id: \"%s\", name: \"%s\">" % \
            (type(self).__name__, self.id, self.name)

    def __setattr__(self, key, value):
        if key not in ('id', 'name', '_items'):
            self.fields[key] = value
        super(DataPipelineObject, self).__setattr__(key, value)

    def __iter__(self):

        def iterhelper(keyvalue):
            key, value = keyvalue
            if isinstance(value, list):
                for val in value:
                    for item in iterhelper((key, val)):
                        yield item
            elif isinstance(value, DataPipelineObject):
                yield { 'key' : key, 'refValue' : value.id }
            elif isinstance(value, DataPipelineParameter):
                yield { 'key' : key, 'stringValue' : "#{%s}" % value.id }
            elif isinstance(value, bool):
                yield { 'key' : key, 'stringValue' : str(value).lower() }
            elif value is not None:
                yield { 'key' : key, 'stringValue' : str(value) }

        yield 'id',     self.id
        yield 'name',   self.name
        yield 'fields', list(itertools.chain(*map(iterhelper, self.fields.iteritems())))

    @property
    def fields(self):
        try:
            return self._items
        except AttributeError:
            self._items = dict()
            return self._items


class TypedDataPipelineObject(DataPipelineObject):
    def __init__(self, **fields):
        type_ = getattr(self, 'TYPE_NAME', type(self).__name__)
        fields.setdefault('type', type_)
        super(TypedDataPipelineObject, self).__init__(**fields)


class Schedule(TypedDataPipelineObject): pass


class RunnableObject(TypedDataPipelineObject): pass



""" DataPipeline Resources. """


from . import base


__all__ = [
    'Resource',
    'Ec2Resource',
    'EmrCluster']


class Resource(base.TypedDataPipelineObject): pass


class Ec2Resource(Resource): pass


class EmrCluster(Resource): pass



""" DataPipeline Parameters. """


from . import base


__all__ = [
    'String',
    'Integer',
    'Double',
    'AwsS3ObjectKey']


class String(base.TypedDataPipelineParameter): pass


class Integer(base.TypedDataPipelineParameter): pass


class Double(base.TypedDataPipelineParameter): pass


class AwsS3ObjectKey(base.TypedDataPipelineParameter):
    TYPE_NAME = 'AWS::S3::ObjectKey'



""" DataPipeline Actions. """


from . import base


__all__ = [
    'Action',
    'SnsAlarm',
    'Terminate']


class Action(base.TypedDataPipelineObject): pass


class SnsAlarm(Action): pass


class Terminate(Action): pass



""" Pline DataPipeline Helper. """


from . import base
from . import actions
from . import activities
from . import data_formats
from . import data_nodes
from . import databases
from . import keywords
from . import parameters
from . import preconditons
from . import resources
from . import utils
from .base import Schedule
from .pipeline import Pipeline


__author__ = 'amancevice'
__version__ = '0.4.3'



""" DataPipeline. """


import collections
import boto3
import botocore
from . import base, exceptions, keywords


__all__ = ['Pipeline']


class Pipeline(object):
    def __init__(self, name, unique_id, desc=None, region='us-east-1', pipeline_id=None):
        self.name        = name
        self.unique_id   = unique_id
        self.desc        = desc
        self.region_name = region
        self.pipeline_id = pipeline_id
        self.objects     = PipelineCollection()
        self.parameters  = PipelineCollection()
        self._client     = None

    @property
    def client(self):
        if self._client is None:
            self.connect()
        return self._client

    def payload(self, pipeline_id=None):
        """ Pipeline payload. """
        subdict = lambda dct,keys: { k:dct[k] for k in keys if k in dct }
        values  = lambda x: subdict(x, ('id', 'stringValue'))
        objects = lambda x: subdict(x, ('id', 'attributes'))
        return {
            'pipelineId'       : pipeline_id or self.pipeline_id,
            'pipelineObjects'  : list(self.objects),
            'parameterValues'  : map(values, self.parameters),
            'parameterObjects' : map(objects, self.parameters) }

    @staticmethod
    def definition(schedule, **kwargs):
        """ Initialize the 'Default' pipeline definition object. """
        kwargs.setdefault('scheduleType',        keywords.scheduleType.cron)
        kwargs.setdefault('failureAndRerunMode', keywords.failureAndRerunMode.CASCADE)
        kwargs.setdefault('role',                'DataPipelineDefaultRole')
        kwargs.setdefault('resourceRole',        'DataPipelineDefaultResourceRole')
        kwargs['schedule'] = schedule
        return base.DataPipelineObject(name='Default', id='Default', **kwargs)

    def connect(self, *args, **kwargs):
        """ Connect to AWS region.

            Arguments:
                args   (tuple): Arguments to pass to boto3.client() after 'datapipeline'
                kwargs (dict):  Keyword arguments to pass to boto3.client()
                                excluding service_name='datapipeline'

            Returns:
                A handle to the boto connection. """
        kwargs.setdefault('region_name', self.region_name)
        self._client = boto3.client('datapipeline', *args, **kwargs)
        return self._client

    def activate(self, **kwargs):
        """ Activate pipeline. """
        assert self.pipeline_id is not None, "pipeline_id is None"
        try:
            return self.client.activate_pipeline(pipelineId=self.pipeline_id, **kwargs)
        except botocore.exceptions.ClientError as err:
            raise exceptions.ClientError(err)

    def update(self):
        """ Update pipeline definition with self.objects. """
        assert self.pipeline_id is not None, "pipeline_id is None"
        try:
            return self.client.put_pipeline_definition(**self.payload())
        except botocore.exceptions.ClientError as err:
            raise exceptions.ClientError(err)

    def validate(self):
        """ Validate the pipeline definition. """
        assert self.pipeline_id is not None, "pipeline_id is None"
        try:
            return self.client.validate_pipeline_definition(**self.payload())
        except botocore.exceptions.ClientError as err:
            raise exceptions.ClientError(err)

    def create(self):
        """ Create pipeline and set pipeline_id. """
        try:
            response = self.client.create_pipeline(
                name        = self.name,
                uniqueId    = self.unique_id,
                description = self.desc)
        except botocore.exceptions.ClientError as err:
            raise exceptions.ClientError(err)
        self.pipeline_id = response['pipelineId']
        if any(self.objects) or any(self.parameters):
            self.update()
        return response

    def add(self, *objects):
        """ Add an object to the pipeline. """
        return self.objects.add(*objects)

    def add_param(self, *params):
        """ Add a parameter to the pipeline. """
        return self.parameters.add(*params)


class PipelineCollection(collections.MutableSet):
    def __init__(self, *items):
        self._collection = set()
        for item in items:
            self.add(item)

    def __contains__(self, item):
        return item in self.collection

    def __iter__(self):
        return iter(map(dict, self.collection))

    def __len__(self):
        return len(self.collection)

    def add(self, *items):
        """ Add one or more items to the collection. """
        for item in items:
            self.collection.add(item)

    def discard(self, *items):
        """ Discard one or more items from the collection. """
        for item in items:
            self.collection.discard(item)

    @property
    def collection(self):
        """ Collection component. """
        return self._collection



""" DataPipeline Activities. """


from . import base


__all__ = [
    'Activity',
    'CopyActivity',
    'EmrActivity',
    'HiveActivity',
    'HiveCopyActivity',
    'PigActivity',
    'RedshiftCopyActivity',
    'ShellCommandActivity',
    'SqlActivity']


class Activity(base.RunnableObject): pass


class CopyActivity(Activity): pass


class EmrActivity(Activity): pass


class HiveActivity(Activity): pass


class HiveCopyActivity(Activity): pass


class PigActivity(Activity): pass


class RedshiftCopyActivity(Activity): pass


class ShellCommandActivity(Activity): pass


class SqlActivity(Activity): pass



""" DataPipeline Preconditions. """


from . import base


__all__ = [
    'Precondition',
    'DynamoDBDataExists',
    'DynamoDBTableExists',
    'Exists',
    'S3KeyExists',
    'S3PrefixNotEmpty',
    'ShellCommandPrecondition']


class Precondition(base.TypedDataPipelineObject): pass


class DynamoDBDataExists(Precondition): pass


class DynamoDBTableExists(Precondition): pass


class Exists(Precondition): pass


class S3KeyExists(Precondition): pass


class S3PrefixNotEmpty(Precondition): pass


class ShellCommandPrecondition(Precondition): pass



""" DataPipeline Utils. """


__all__ = ['ShellCommand']


class ShellCommand(object):
    def __init__(self, *args):
        self._commands = list(args)

    def __repr__(self):
        if len(self._commands) == 1:
            return self._commands[0]
        return ";\\\n".join(self._commands)

    def __str__(self):
        return repr(self)

    def append(self, *args):
        for arg in args:
            self._commands.append(arg)



""" DataPipeline Keywords. """


__all__ = [
    'actionOnResourceFailure',
    'actionOnTaskFailure',
    'failureAndRerunMode',
    's3EncryptionType',
    'scheduleType',
    'startAt']


def _keywords(*values):
    return type('KEYWORD', (), dict([('__slots__', values)] + [(v,v) for v in values]))


actionOnResourceFailure = _keywords('retryAll', 'retryNone')
actionOnTaskFailure     = _keywords('continue', 'terminate')
failureAndRerunMode     = _keywords('CASCADE', 'NONE')
s3EncryptionType        = _keywords('SERVER_SIDE_ENCRYPTION', 'NONE')
scheduleType            = _keywords('timeseries', 'cron', 'ondemand')
startAt                 = _keywords('FIRST_ACTIVATION_DATE_TIME')



""" DataPipeline Databases. """


from . import base


__all__ = [
    'Database',
    'JdbcDatabase',
    'RdsDatabase',
    'RedshiftDatabase']


class Database(base.TypedDataPipelineObject): pass


class JdbcDatabase(Database): pass


class RdsDatabase(Database): pass


class RedshiftDatabase(Database): pass



""" DataPipeline Data Formats. """


from . import base


__all__ = [
    'DataFormat',
    'CSV',
    'Custom',
    'DynamoDBDataFormat',
    'DynamoDBExportDataFormat',
    'RegEx',
    'TSV']


class DataFormat(base.TypedDataPipelineObject): pass


class CSV(DataFormat): pass


class Custom(DataFormat): pass


class DynamoDBDataFormat(DataFormat): pass


class DynamoDBExportDataFormat(DataFormat): pass


class RegEx(DataFormat): pass


class TSV(DataFormat): pass



""" DataPipeline Data Nodes. """


from . import base


__all__ = [
    'DataNode',
    'DynamoDBDataNode',
    'MySqlDataNode',
    'RedshiftDataNode',
    'S3DataNode',
    'SqlDataNode']


class DataNode(base.RunnableObject): pass


class DynamoDBDataNode(DataNode): pass


class MySqlDataNode(DataNode): pass


class RedshiftDataNode(DataNode): pass


class S3DataNode(DataNode): pass


class SqlDataNode(DataNode): pass
